//cohomology.m 4/24/09

freeze;

import "findcells.m":
    GetSign;
import "defs.m":
    HomologyRec;
import "setlevel.m":
    brm,
    SetLevel;

function number_of_orbits(orb)
    return Maximum(orb`orbit_type);
end function;

function get_column(M,GLtype2,Gamma0type2);
    //returns the column in the boundary matrix associated to the GLtype
    //and Gamma0type.  Does not throw out non-orientable sharblies.
    //Just adds them to the relations.
    GLstd2:=M`ModFrmData`two_poly[GLtype2];
    orbits2:=M`LevelData`two_orbits[GLtype2];
    gamma0:=orbits2`gamma_list[Gamma0type2];
    boundary:=GLstd2`GL_facet_types;
    sofar:=[0:i in [1..M`LevelData`total_number_of_one_orbits]];
    
    //We know the boundary of the standard 2 polytopes.  We get the
    //boundaries of the Gamma0 classes of 2-polytopes by translation.
    for L in boundary do
	//gamma0 moves standard 2-poly to here.
	mover:=L[3];//moves std edge to here
	sgn:=L[2];
	GLtype1:=L[1];
	i:=brm(M,gamma0*mover);
	Gamma0type1:=M`LevelData`one_orbits[GLtype1]`orbit_type[i];
		
	onumber:=M`LevelData`one_orbits[GLtype1]`orbit_number_list[i];
	totsgn:=sgn*onumber;
	j:=M`LevelData`one_orbits[GLtype1]`entry[Gamma0type1];
	sofar[j]:=totsgn + sofar[j];
        vprintf Bianchi,4: "So far %o.\n",
	Support(Vector(sofar));
    end for;
    vprintf Bianchi,4: "This relation involves edges %o.\n",
	Support(Vector(sofar));
    return sofar;
end function;


function relations(M)
    //Assumes F and M`LevelData have already been set up correctly
    vprint Bianchi, 2:"Computing relations and non-orientable cells.";
    orbits2:=M`LevelData`two_orbits;
    orbits1:=M`LevelData`one_orbits;
    L:=[];
    for GLtype2 in [1..#orbits2] do
	orbit:=orbits2[GLtype2];
	for Gamma0type2 in [1..orbit`total_number] do
	    col:=get_column(M,GLtype2,Gamma0type2);
	    Append(~L,col);
	end for;
    end for;
    //the above computes the relations coming from the facets of the
    //polytopes.  Now we still must introduce the relations coming from
    //non-orientable orbits.
    nonoL:=[];
    for GLtype1 in [1..#orbits1] do
	edge:=orbits1[GLtype1];
	noedge_index:=[edge`entry[i]: i in edge`non_orientable_orbits];
	non_orientables:=[[0: i in [1..M`LevelData`total_number_of_one_orbits]]:
	    j in [1..#noedge_index]];
	for i in [1..#noedge_index] do
	    non_orientables[i][noedge_index[i]]:=1;
	    vprintf Bianchi, 4:
		"edge number %o is non-orientable.\n", noedge_index[i];
	end for;
	nonoL:=	nonoL cat non_orientables;
    end for;
    /*
    //Test if ok that there are no nonorientable orbits.
    nonoL:=[];
    */
    L:=L cat nonoL;
    vprintf Bianchi, 3:
	"     There are %o relations and %o non-orientable edges.\n",#L,#nonoL;
    return L, noedge_index;
end function;


procedure HomologyData(M)
    //Attaches a record of the homology
    SetLevel(M);
    Coeff:=M`homology_coefficients;
    if not assigned  M`LevelData`homology`relations then 
	rel:=relations(M);
	M`LevelData`homology`relations:=rel;
    end if;
    M`LevelData`homology`coefficient_ring:=Coeff;
    if not assigned M`LevelData`homology`zero_sharbly_space then 
	zss:=RSpace(Coeff,M`LevelData`total_number_of_one_orbits);
	M`LevelData`homology`zero_sharbly_space:=zss;
    end if;
    if not assigned M`LevelData`homology`H then 
	vprint Bianchi, 2:"Computing H^2.";
	H,f:=quo<zss|rel>;
	vprintf Bianchi,2:"H^2 is %o.\n", H;
	M`LevelData`homology`H:=H;
	M`LevelData`homology`ToH:=f;
	M`Full_Dimension:=Rank(H);
    end if;
end procedure;
