
freeze;

Z:=Integers(); Q:=Rationals();

intrinsic KroneckerCharacter(D::FldRatElt) -> GrpDrchElt
{The Kronecker character (D/n)}
 require D ne 0: "Cannot have KroneckerCharacter of 0";
 return KroneckerCharacter(Numerator(D)*Denominator(D)); end intrinsic;

intrinsic IdentifyKroneckerCharacter(chi::GrpDrchElt) -> RngIntElt
{Given a real Dirichlet character (over Q), identify it as a Kronecker char }
 chi:=AssociatedPrimitiveCharacter(chi);
 require Order(chi) le 2: "Must be quadratic";
 if Order(chi) eq 1 then return 1; end if; C:=Conductor(chi);
 return chi eq KroneckerCharacter(C) select C else -C; end intrinsic;

intrinsic ArtinRepresentationQuadratic(d::FldRatElt) -> ArtRep
{Return the Artin representation corresponding to the character given by d}
 require d ne 0: "Discriminant cannot be 0";
 if IsSquare(d) then return ArtinRepresentations(Rationals())[1]; end if;
 x:=PolynomialRing(Rationals()).1;
 return ArtinRepresentations(NumberField(x^2-d))[2]; end intrinsic;

intrinsic ArtinRepresentationQuadratic(d::RngIntElt) -> ArtRep {"}//"
 return ArtinRepresentationQuadratic(Q!d); end intrinsic;

////////////////////////////////////////////////////////////////////////

ArtRepQuad:=ArtinRepresentationQuadratic;

function perm_char(F) // permutation character of an algebra
 return &+[PermutationCharacter(NumberField(f[1])) : f in Factorization(F)];
 end function;

function get_art_rep(X,t)
 if #X`Glist eq 3 then a:=-X`Glist[1]; b:=-X`Glist[2]; c:=Gcd(a,b);
  M:=MValue(X); u:=1/M/t; x:=PolynomialRing(Rationals()).1;
  return perm_char(x^a*(1-x)^b-u)-perm_char(x^c-u); end if; H:=Twist(X);
 if #H`Glist eq 3 then M:=MValue(H);
  if IsOdd(#X`alpha) then u:=1/t; tw:=-M; else u:=t; tw:=M; end if;
   _,tw:=Valuation(tw,2); tw:=Numerator(tw)*Denominator(tw);
   tw:=Sign(tw)*&*[Z|f[1]^(f[2] mod 2) : f in Factorization(tw)];
   return ArtinRepresentation(H,u)*ArtRepQuad(tw*t); end if;
 return false; end function; // could automate via GammaList, but for twisting

procedure art_rep_check(R,H,t) P:=PrimesInInterval(2,100);
 Q:=Set(P) diff Set(PrimeDivisors(Numerator(t)));
 Q:=Set(Q) diff Set(PrimeDivisors(Denominator(t)));
 Q:=Set(Q) diff Set(PrimeDivisors(Numerator(t-1)));
 Q:=Set(Q) diff
    Set(PrimeDivisors(LCM([Denominator(x) : x in H`alpha cat H`beta])));
 A:=[EulerFactor(H,t,p) : p in Q]; B:=[EulerFactor(R,p : R:=Z) : p in Q];
 assert A eq B; end procedure;

intrinsic ArtinRepresentation(X::HypGeomData,t::FldRatElt:Check:=true)->ArtRep
{Given a weight 0 hypergeometric motive specialised at t, return the
 corresponding Artin representation (if known). Checks primes up to 100.}
 require t ne 0 and t ne 1: "t must be neither 0 nor 1";
 require X`weight eq 0: "Hypergeometric motive must be weight 0";
 A:=get_art_rep(X,t);
 if Type(A) eq BoolElt then
  require false: "Hypergeometric data not known as an Artin rep"; end if;
 if Check then art_rep_check(A,X,t); end if; return A; end intrinsic;

intrinsic ArtinRepresentation(X::HypGeomData,t::RngIntElt:Check:=true)->ArtRep
{"} return ArtinRepresentation(X,Rationals()!t:Check:=Check); end intrinsic;//"

////////////////////////////////////////////////////////////////////////

function hypergeom_crvell_deg2(X,t) // should rewrite this
 if X`beta eq [0,0] then // weight 1 cases
  if X`alpha eq [1/2,1/2] then a:=1/(16*t); // self twist
   E:=EllipticCurve([1,-2*a,0,a^2,0]); return E; end if;
  if X`alpha eq [1/3,2/3] then a:=1/(27*t);
   E:=EllipticCurve([1,0,a,0,0]); return E; end if;
  if X`alpha eq [1/4,3/4] then a:=1/(64*t);
   E:=EllipticCurve([1,0,0,a,0]); return E; end if;
  if X`alpha eq [1/6,5/6] then a:=1/(432*t);
   E:=EllipticCurve([1,0,0,0,-a]); return E; end if; end if;
 if X`alpha eq [1/2,1/2] then // weight 1 cases
  if X`beta eq [1/3,2/3] then a:=27/(16*t);
   E:=EllipticCurve([0,-a,-a,0,0]); return E; end if;
  if X`beta eq [1/4,3/4] then a:=4/t;
   E:=EllipticCurve([0,-a,0,a,0]); return E; end if;
  if X`beta eq [1/6,5/6] then a:=27/t;
   E:=EllipticCurve([0,-a/4,0,a/2,-a/4]); return E; end if; end if;
 if X`alpha eq [1/3,2/3] and X`beta eq [1/4,3/4] then a:=64/(27*t);
  E:=EllipticCurve([0,0,-a,-a,0]); return E; end if;
 if X`alpha eq [1/3,2/3] and X`beta eq [1/6,5/6] then a:=16/t; // self-twist
  E:=EllipticCurve([0,0,0,-3*a,4*a+a^2/4]); return E; end if;
 if X`alpha eq [1/4,3/4] and X`beta eq [1/6,5/6] then a:=27/(4*t);
  E:=EllipticCurve([0,0,0,-a,a]); return E; end if; end function;

function ec_gen(X,t) _,d:=IsPrimitive(X); H:=PrimitiveData(X); F:=Parent(t);
 u:=1/(X`Mvalue*t); P:=PolynomialRing(F); y:=P.1; FAC:=Factorization(y^d-u);
 if F cmpeq Q then Ks:=[* ext<F|f[1] : DoLinearExtension> : f in FAC *];
 else Ks:=[* ext<F|f[1]> : f in FAC *]; end if;
 if H`cycA eq [6] and H`cycB eq [1,1] then
  Es:=[* EllipticCurve([1,0,0,0,-K.1]) : K in Ks *]; end if;
 if H`cycA eq [4] and H`cycB eq [1,1] then
  Es:=[* EllipticCurve([1,0,0,K.1,0]) : K in Ks *]; end if;
 if H`cycA eq [3] and H`cycB eq [1,1] then
  Es:=[* EllipticCurve([1,0,K.1,0,0]) : K in Ks *]; end if;
 if H`cycA eq [2,2] and H`cycB eq [1,1] then
  Es:=[* EllipticCurve([1,K.1,K.1,0,0]) : K in Ks *]; end if;
 if H`cycA eq [4] and H`cycB eq [6] then
  Es:=[* EllipticCurve([0,0,0,-K.1,K.1]) : K in Ks *]; end if;
 if H`cycA eq [3] and H`cycB eq [6] then
  Es:=[* EllipticCurve([0,0,0,-3*K.1,4*K.1+K.1^2/4]) : K in Ks *]; end if;
 if H`cycA eq [2,2] and H`cycB eq [6] then
  Es:=[* EllipticCurve([0,-K.1,0,8*K.1,-16*K.1]) : K in Ks *]; end if;
 if H`cycA eq [3] and H`cycB eq [4] then
  Es:=[* EllipticCurve([0,0,-K.1,-K.1,0]) : K in Ks *]; end if;
 if H`cycA eq [2,2] and H`cycB eq [4] then
  Es:=[* EllipticCurve([0,-K.1,0,K.1,0]) : K in Ks *]; end if;
 if H`cycA eq [2,2] and H`cycB eq [3] then
  Es:=[* EllipticCurve([0,-K.1,K.1,0,0]) : K in Ks *]; end if;
 assert assigned Es; return #Es eq 1 select Es[1] else Es; end function;

procedure check_ec(C,H,t) P:=PrimesInInterval(2,50); // curtail deg in Euler?
 if Type(C) ne List then C:=[* C *]; end if; // hack
 Q:=Set(P) diff Set(PrimeDivisors(Numerator(t)));
 Q:=Set(Q) diff Set(PrimeDivisors(Denominator(t)));
 Q:=Set(Q) diff Set(PrimeDivisors(Numerator(t-1)));
 Q:=Set(Q) diff
    Set(PrimeDivisors(LCM([Denominator(x) : x in H`alpha cat H`beta])));
 A:=[EulerFactor(H,t,p) : p in Q]; B:=[&*[EulerFactor(c,p) : c in C] : p in Q];
 assert A eq B; end procedure;

intrinsic
EllipticCurve(X::HypGeomData,t::RngIntElt : Check:=false) -> CrvEll {"}//"
 return EllipticCurve(X,Rationals()!t : Check:=Check); end intrinsic;

intrinsic EllipticCurve(X::HypGeomData,t::FldRatElt : Check:=false) -> CrvEll
{Given suitable hypergeometric data (of weight 1 and imprimitivity=degree/2),
 return the associated elliptic curve specialized at t.
 Returns a list of curves when the algebra splits.}
 require t ne 0 and t ne 1: "t must be neither 0 nor 1";
 require Weight(X) eq 1: "Weight must be 1"; _,g:=IsPrimitive(X);
 require Degree(X) eq 2*g: "Imprimitivity must be half the degree";
 if Degree(X) eq 2 and Weight(X) eq 1 then
  return hypergeom_crvell_deg2(X,t); end if;
 E:=ec_gen(X,t); if Check then check_ec(E,X,t); end if; return E;
 end intrinsic;

intrinsic EllipticCurve(X::HypGeomData) -> CrvEll
{Given suitable hypergeometric data (of weight 1 and imprimitivity=degree/2),
 return the associated elliptic curve (over a function field).}
 require Weight(X) eq 1: "Weight must be 1"; _,g:=IsPrimitive(X);
 require Degree(X) eq 2*g: "Imprimitivity must be half the degree";
 if Degree(X) eq 2 and Weight(X) eq 1 then
  return hypergeom_crvell_deg2(X,FunctionField(Rationals()).1); end if;
 return ec_gen(X,FunctionField(Rationals()).1); end intrinsic;

////////////////////////////////////////////////////////////////////////

intrinsic HyperellipticCurve(S::SeqEnum) -> CrvHyp
{Given a sequence of coefficients, construct a hyperelliptic curve}
 return HyperellipticCurve(Polynomial(Reverse(S))); end intrinsic;

procedure check_hc(C,H,t) P:=PrimesInInterval(2,50);
 Q:=Set(P) diff Set(PrimeDivisors(Numerator(t)));
 Q:=Set(Q) diff Set(PrimeDivisors(Denominator(t)));
 Q:=Set(Q) diff Set(PrimeDivisors(Numerator(t-1)));
 Q:=Set(Q) diff
    Set(PrimeDivisors(LCM([Denominator(x) : x in H`alpha cat H`beta])));
 A:=[EulerFactor(H,t,p) : p in Q]; B:=[EulerFactor(C,p) : p in Q];
 assert A eq B; end procedure;

function hyperell_curve(X,t) // 18 times 2 cases known
 a:=Sort(X`cycA); b:=Sort(X`cycB); M:=X`Mvalue; u:=1/(M*t);
 if a eq [3,6] and b eq [1,1,4] then // case 1
  return HyperellipticCurve([1,2,1,0,0,0,-4*u]); end if;
 if a eq [2,2,4] and b eq [3,6] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),-t); end if;
 if a eq [5] and b eq [1,1,4] then // case 2
  return HyperellipticCurve([1,2,1,0,0,-4*u,-4*u]); end if;
 if a eq [2,2,4] and b eq [10] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),-5*t); end if;
 if a eq [8] and b eq [1,1,3] then // case 3
  return HyperellipticCurve([4,1,0,0,4*u,0]); end if;
 if a eq [2,2,6] and b eq [8] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),-3*t); end if;
 if a eq [5] and b eq [1,1,3] then // case 4
  return HyperellipticCurve([1,2,1,0,0,4*u,0]); end if;
 if a eq [2,2,6] and b eq [10] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),-15*t); end if;
 if a eq [4,4] and b eq [1,1,3] then // case 5
  return HyperellipticCurve([4,1,8*u,0,4*u^2,0]); end if;
 if a eq [4,4] and b eq [2,2,6] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),-3*t); end if;
 if a eq [2,2,4] and b eq [1,1,3] then // case 6
  return HyperellipticCurve([1,2,1,0,4*u,4*u,0]); end if;
 if a eq [2,2,6] and b eq [1,1,4] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),3*t); end if;
 if a eq [5] and b eq [1,1,2,2] then // case 7
  return HyperellipticCurve([4,1,0,-2*u,0,u^2]); end if;
 if a eq [10] and b eq [1,1,2,2] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),5*t); end if;
 if a eq [3,4] and b eq [1,1,2,2] then // case 8
  return HyperellipticCurve([4,1,-4*u,-2*u,0,u^2]); end if;
 if a eq [4,6] and b eq [1,1,2,2] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),3*t); end if;
 if a eq [5] and b eq [8] then // case 9
  return HyperellipticCurve([1,4,0,0,0,-4*u,0]); end if;
 if a eq [8] and b eq [10] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),5*t); end if;
 if a eq [5] and b eq [3,6] then // case 10
  return HyperellipticCurve([1,0,0,0,0,-4*u,4*u]); end if;
 if a eq [3,6] and b eq [10] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),5*t); end if;
 if a eq [5] and b eq [2,2,6] then // case 11
  return HyperellipticCurve([4,0,0,-4*u,0,u^2]); end if;
 if a eq [10] and b eq [1,1,3] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),-15*t); end if;
 if a eq [3,3] and b eq [2,2,6] then // case 12
  return HyperellipticCurve([1,4,0,-2*u,-4*u,0,u^2]); end if;
 if a eq [6,6] and b eq [1,1,3] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),-3*t); end if;
 if a eq [4,4] and b eq [5] then // case 13
  return HyperellipticCurve([4,0,-8*u,0,4*u^2,u^2]); end if;
 if a eq [4,4] and b eq [10] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),5*t); end if;
 if a eq [3,4] and b eq [5] then // case 14
  return HyperellipticCurve([4,0,-4*u,0,0,u^2]); end if;
 if a eq [4,6] and b eq [10] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),15*t); end if;
 if a eq [2,2,4] and b eq [5] then // case 15
  return HyperellipticCurve([1,0,0,0,-4*u,4*u,0]); end if;
 if a eq [10] and b eq [1,1,4] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),-5*t); end if;
 if a eq [3,3] and b eq [5] then // case 16
  return HyperellipticCurve([1,4,0,-2*u,0,0,u^2]); end if;
 if a eq [6,6] and b eq [10] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),5*t); end if;
 if a eq [3,3] and b eq [2,2,4] then // case 17
  return HyperellipticCurve([1,0,0,2*u,-4*u,0,u^2]); end if;
 if a eq [6,6] and b eq [1,1,4] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),-t); end if;
 if a eq [3,3] and b eq [1,1,2,2] then // case 18
  return HyperellipticCurve([1,2,1,2*u,-2*u,0,u^2]); end if;
 if a eq [6,6] and b eq [1,1,2,2] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),t); end if;
////////////////////////////////////////////////////////////////////////
 if a eq [8] and b eq [1,1,4] then // case C1
  return HyperellipticCurve([4,1,0,0,-16*u,-4*u]); end if;
 if a eq [2,2,4] and b eq [8] then
  return QuadraticTwist(hyperell_curve(Twist(X),1/t),-t); end if;
 if a eq [4,6] and b eq [1,1,3] then // case C2
  return HyperellipticCurve([4/u,0,-16,1,0,-4*u]); end if;
 if a eq [3,4] and b eq [2,2,6] then
  return QuadraticTwist(hyperell_curve(Twist(X),t),-t); end if;
 if a eq [2,2,6] and b eq [1,1,3] then // case C3
  return HyperellipticCurve([4,1,0,16*u,4*u,0]); end if; // self-twist
 if a eq [5] and b eq [10] then // case D1
  return HyperellipticCurve([1,2,0,0,0,u/8,u/4]); end if; // self-twist
 return false; end function;

intrinsic HyperellipticCurve(H::HypGeomData) -> CrvHyp
{Given suitable hypergeometric data, return the associated
 hyperelliptic curve (over a function field). If not known, return false.}
 require Degree(H) eq 4 and Weight(H) eq 1: "Degree must be 4 and weight 1";
 t:=FunctionField(Rationals()).1; return hyperell_curve(H,t); end intrinsic;

intrinsic HyperellipticCurve
  (H::HypGeomData,t::FldRatElt : Check:=true) -> CrvHyp
{Given suitable hypergeometric data, return the associated
 hyperelliptic curve at t. If not known, return false.}
 require Degree(H) eq 4 and Weight(H) eq 1: "Degree must be 4 and weight 1";
 require t ne 0 and t ne 1: "t must be neither 0 nor 1";
 C:=hyperell_curve(H,t); if Type(C) eq BoolElt then return false; end if;
 if Check then check_hc(C,H,t); end if; return C; end intrinsic;

intrinsic HyperellipticCurve
  (H::HypGeomData,t::RngIntElt : Check:=true) -> CrvHyp {"}//"
 return HyperellipticCurve(H,Q!t : Check:=Check); end intrinsic;

intrinsic Identify(H::HypGeomData,t::FldRatElt) -> Any
{Try to identify hypergeometric data at a given t as an associated object.
 Returns false if unable.}
 require t ne 0 and t ne 1: "t must be neither 0 nor 1";
 if Weight(H) gt 1 then return false; end if;
 if Weight(H) eq 0 then return get_art_rep(H,t); end if;
 if Weight(H) eq 1 then _,g:=IsPrimitive(H);
  if 2*g eq Degree(H) then return EllipticCurve(H,t); end if;
  if Degree(H) ne 4 then return false; end if;
  T:=HyperellipticCurve(H,t); return T; end if; end intrinsic;

intrinsic Identify(H::HypGeomData,t::RngIntElt) -> Any {"}//" t a RngElt?
 return Identify(H,Q!t); end intrinsic;

////////////////////////////////////////////////////////////////////////

function hgm_curve(X,t) G:=[x: x in GammaList(X)];
 F:=FieldOfFractions(Parent(t)); A:=AffineSpace(F,2); x:=A.1; y:=A.2;
 u:=1/(MValue(X)*t); if #G gt 6 then return false; end if;
 if #G eq 4 then
  if Sign(G[4]) ne Sign(G[1]) then G:=[Abs(x) : x in G]; assert G[4] gt 0;
   f:=x^G[1]*y^G[2]*(1-x-y)^G[3]-u; return Curve(A,f); end if;
  G:=[Abs(x) : x in G]; f:=x^G[4]*(1-x)^G[1]-y^G[3]*(1-y)^G[2]*u;
  return Curve(A,f); end if;
 for i in [1..5],j in [i+1..5] do
  if G[i]+G[j]+G[#G] eq 0 then ei:=G[i]; ej:=G[j];
   a:=Abs(G[i]); b:=Abs(G[j]); s1:=Sign(G[#G]);
   Exclude(~G,ei); Exclude(~G,ej); Exclude(~G,G[#G]);
   c:=Abs(G[1]); d:=Abs(G[2]); s2:=Sign(G[3]);
   if s1 eq -1 then assert s2 eq +1;
    f:=x^c*(1-x)^d-y^a*(1-y)^b*u; return Curve(A,f); end if;
   if s2 eq -1 then
    f:=x^a*(1-x)^b-y^c*(1-y)^d*u; return Curve(A,f); end if;
   f:=x^a*(1-x)^b*y^c*(1-y)^d-u; return Curve(A,f); end if; end for;
 return false; end function;

function canonical_scheme(X,t)
 G:=X`Glist; Gn:=[-x : x in G | x lt 0]; Gp:=[x : x in G | x gt 0];
 A:=AffineSpace(Parent(t),#Gp+#Gn);
 f1:=&+[A.i : i in [1..#Gn]]-1; f2:=&+[A.(#Gn+i) : i in [1..#Gp]]-1;
 f3:=&*[A.i^Gn[i] : i in [1..#Gn]]-
     &*[A.(#Gn+i)^Gp[i] : i in [1..#Gp]]/MValue(X)/t;
 return Scheme(A,[f1,f2,f3]); end function;

intrinsic CanonicalCurve(X::HypGeomData) -> Crv
{Returns a canonical plane curve (possibly reducible) associated to 
 the given hypergeometric data, or false if this is not possible}
 return hgm_curve(X,FunctionField(Rationals()).1); end intrinsic;

intrinsic CanonicalCurve(X::HypGeomData,t::FldRatElt) -> Crv
{Returns a canonical plane curve (possibly reducible) associated to 
 the given hypergeometric data specialised at t,
 or false if this is not possible}
 return hgm_curve(X,t); end intrinsic;

intrinsic CanonicalCurve(X::HypGeomData,t::RngIntElt) -> Crv
{"} return hgm_curve(X,Q!t); end intrinsic; //"

intrinsic CanonicalScheme(X::HypGeomData) -> Sch
{The canonical scheme associated to given hypergeometric data,
 returned over a function field}
 return canonical_scheme(X,FunctionField(Rationals()).1); end intrinsic;

intrinsic CanonicalScheme(X::HypGeomData,t::FldRatElt) -> Sch
{The canonical scheme associated to given hypergeometric data,
  specialized at a rational t.}
 return canonical_scheme(X,t); end intrinsic;

intrinsic CanonicalScheme(X::HypGeomData,t::RngIntElt) -> Sch
{"} return CanonicalScheme(X,Q!t); end intrinsic; //"
