freeze;

//freeze;

Z := IntegerRing();

function Pivots(X)
    // Pivot columns of X
    E := EchelonForm(X);
    P := [];
    for i := 1 to Nrows(E) do
        S := Support(E[i]);
        if #S gt 0 then
            Append(~P, Min(S));
        end if;
    end for;
    return P;
end function;

intrinsic UnimodularExtension(X::Mtrx) -> Mtrx
{Given a basis matrix X with trivial elementary divisors, return a matrix
U such that the vertical join of X with U is unimodular}

    require BaseRing(X) cmpeq Z: "Matrix must be integral";
    // Don't bother to check ED

    vprint Isolution: "Get Smith with right inverse"; vtime Isolution:
    S, U := SmithForm(X: RightInverse);

    r := Nrows(X);
    c := Ncols(X);
    assert Nrows(U) eq c;
    U := RowSubmatrixRange(U, r + 1, c);
    return U;
    // J := VerticalJoin(X, U); return J;

end intrinsic;

intrinsic LatticeBasisMatrix(X::Mtrx) -> Mtrx, Mtrx, SeqEnum, Mtrx
{Return a basis matrix B of the lattice spanned by the rows of X, with
transformation T so that T*X = B, the list of indices for the rows of
X that are actually used, and the matrix of those rows from X}

    require BaseRing(X) cmpeq Z: "Matrix must be integral";

    vprint Isolution: "Get ElementaryDivisors"; vtime Isolution:
    E := ElementaryDivisors(X);
    rank := #E;
    EP := &*E;

    USE_KER := true;

    vprint Isolution: "Input ED:", {* x: x in E *};
    vprint Isolution: "Input rank:", rank;

    m := Nrows(X);
    if rank eq m then
	return X, MatrixRing(Z, m)!1, [1 .. m], X;
    end if;

    d := E[#E];
    p := 1;
    repeat
	p := NextPrime(p);
    until d mod p ne 0;
    vprint Isolution: "Get pivots mod", p; vtime Isolution:
    piv := Pivots(Matrix(GF(p), Transpose(X)));
    pivs := Set(piv);
    // Remove repeated rows?
    pp := piv cat [i: i in [1 .. m] | i notin pivs];

    last_EY := 0; e := 0;
last_l := rank;
    e := 1;
    e := 5;
    e := 10;

e := 20;
e := 30;

    ecount := 0;
    step := 0;

    if USE_KER then
	det := 0;

	Ypp := pp[1 .. rank];
	Y := RowSubmatrix(X, Ypp);

	vprint Isolution: "Get determinant of initial submatrix";
	//vtime Isolution: det := Determinant(Y);
	vtime Isolution: det := &*ElementaryDivisors(Y);
	vprint Isolution: "Initial quotient:", det/EP;

	/*
	"Det Fact:", Factorization(det);
	"EP fact:", Factorization(EP);
	"Init quot fact:", Factorization(Z!(det/EP));
	*/

	Ldet := det;

	while Ldet ne EP do
	    step +:= 1;

	    l := Min(last_l + e, #pp);
	    ecount +:= e;
	    Ypp := pp[1 .. l];
	    Y := RowSubmatrix(X, Ypp);
//Ypp;
	    vprint Isolution: "\n*** STEP", step;
	    vprintf Isolution: "Quotient: %o\n", Ldet/EP;
	    vprintf Isolution: "rank: %o, last_l: %o, l: %o\n", rank, last_l, l;

	    vprint Isolution: "Get kernel";
	    vtime Isolution: k := EchelonForm(KernelMatrix(ReverseRows(Y)));
	    kr := Nrows(k);
	    assert kr eq Nrows(Y) - rank;
	    sk := ColumnSubmatrix(k, 1, kr);
	    //sk := ReverseColumns(sk);
	    sk := EchelonForm(sk);
	    //"sk:", sk;
	    skd := Diagonal(sk);

	    skd := Reverse(skd);
	    vprint Isolution: "Fixed sk Diag:", skd;

	    ind := [i: i in [1 .. #skd] | skd[i] ne 1];
	    vprint Isolution: "ind:", ind;
	    keep := [1 .. rank] cat [rank + i: i in ind] cat
		[rank + kr + 1 .. #pp];
	    pp := pp[keep];
	    last_l := rank + #ind;
	    Ypp := pp[1 .. last_l];
	    Y := RowSubmatrix(X, Ypp);

	    // "New &*skd fact:", Factorization(&*skd);
	    Ldet := ExactQuotient(det, &*skd);

	    /*
	    vprintf Isolution:
		"Get ElementaryDivisors for %o + %o rows\n",
		rank, Nrows(Y) - rank;

	    vtime Isolution: EY := ElementaryDivisors(Y);
	    vprint Isolution: "EY:", {* x: x in EY *};
	    EPY := &*EY;
	    vprint Isolution: "E quotient:", EPY/EP;
	    if EY eq E then
		break;
	    end if;
	    last_EY := EY;
	    */

	end while;
    else
	while true do
	    step +:= 1;

	    l := Min(last_l + e, #pp);
	    ecount +:= e;
	    Ypp := pp[1 .. l];
	    Y := RowSubmatrix(X, Ypp);
//Ypp;
	    vprint Isolution: "\n*** STEP", step;
	    vprintf Isolution: "rank: %o, last_l: %o, l: %o\n", rank, last_l, l;

	    if USE_KER then
		vprint Isolution: "Get kernel";
		vtime Isolution: k := KernelMatrix(ReverseRows(Y));
		kr := Nrows(k);
		assert kr eq Nrows(Y) - rank;
		sk := ColumnSubmatrix(k, 1, kr);
		//sk := ReverseColumns(sk);
		sk := EchelonForm(sk);
		//"sk:", sk;
		skd := Diagonal(sk);

		skd := Reverse(skd);
		vprint Isolution: "Fixed sk Diag:", skd;

		ind := [i: i in [1 .. #skd] | skd[i] ne 1];
		vprint Isolution: "ind:", ind;
		keep := [1 .. rank] cat [rank + i: i in ind] cat
		    [rank + kr + 1 .. #pp];
		pp := pp[keep];
		last_l := rank + #ind;
		Ypp := pp[1 .. last_l];
		Y := RowSubmatrix(X, Ypp);

		// KP := &*skd; "KP quot:", EP/KP;
	    end if;

	    vprintf Isolution:
		"Get ElementaryDivisors for %o + %o rows\n",
		rank, Nrows(Y) - rank;

	    vtime Isolution: EY := ElementaryDivisors(Y);
	    vprint Isolution: "EY:", {* x: x in EY *};
	    EPY := &*EY;
	    vprint Isolution: "E quotient:", EPY/EP;
	    if EY eq E then
		break;
	    end if;

	    if USE_KER then
		;
	    elif last_EY cmpne 0 and EY eq last_EY then
		vprint Isolution: "No change; remove extra row indices";
		//pp := pp[1 .. l - e] cat pp[l + 1 .. #pp];
//last_l, l;
assert l gt last_l;
		pp := pp[1 .. last_l] cat pp[l + 1 .. #pp];
	    else
		//e +:= 10;
		last_l := l;
	    end if;
	    last_EY := EY;
    //assert ElementaryDivisors(RowSubmatrix(X, pp)) eq E;
	end while;
    end if;

    vprintf Isolution: "Final number of extra rows: %o\n", Nrows(Y) - rank;

    vprint Isolution: "Get kernel"; vtime Isolution:
    N := KernelMatrix(Y);

//N;

    vprint Isolution: "Kernel dim:", Nrows(N);
    if 1 eq 1 then
	vprint Isolution: "Get Echelon of kernel"; vtime Isolution:
	N := EchelonForm(N);
    else
	vprint Isolution: "Get LLL of kernel"; vtime Isolution:
	//N := QuickLLL(N);
	N := LLL(N: Proof := false);
    end if;

    vprint Isolution: "LLL Kernel log norms:", LogNorms(N);

    U := UnimodularExtension(N);
    //delete N;

//"First U:", U;
    //U := QuickLLL(U);
//"Next U:", U;

    return U*Y, U, N, Ypp, Y, pp;

end intrinsic;

intrinsic NHerm(X::Mtrx) -> Mtrx
{Internal; assumes rank eq Ncols(X)?}

// SetVerbose("Isolution",1); "NHerm"; X: Magma;

    n := Ncols(X);
    m := Nrows(X);
    e := 1;
    s := n - e;
    S := ColumnSubmatrix(X, 1, s);

    B, U, N, Ypp, Y := LatticeBasisMatrix(S);
    pp := Ypp cat [i: i in [1 .. m] | i notin Ypps] where Ypps := Set(Ypp);
    X := RowSubmatrix(X, pp);

    k := #Ypp;
    assert k ge s;
    d := k - s;

    C := Submatrix(X, 1, s + 1, k, e);
    UC := U*C;
    NC := N*C;

    //"UC:", Parent(UC); UC;
    //"NC:", Parent(NC); NC;

    SED := ElementaryDivisors(S);
    M := 2*SED[#SED];
    R := IntegerRing(M);
    S := Matrix(R, S);

    PACKED_LIMIT := 16;

    Sd := Density(S);
    if Sd le 0.1 and M le PACKED_LIMIT then

	S := SparseMatrix(S);
	vprintf Isolution: "Get Hermite of S (sparse density %.4o) mod %o\n",
	    Sd, M;
	vtime Isolution: Echelonize(~S);
	S := SparseMatrix(Z, S);
	S := Matrix(S);
	SH := S;
    else
	vprint Isolution: "Get Hermite S mod", M; vtime Isolution:
	S := EchelonForm(S);
	SH := Matrix(Z, EchelonForm(S));
    end if;
    delete S;

    RemoveZeroRows(~SH);

    vprint Isolution: "Get PseudoInverse of SH"; vtime Isolution:
    PI,den := PseudoInverse(Matrix(SH));

    vprint Isolution: "Get product of B and PI"; vtime Isolution:
    P := (B*PI) div den;

    vprint Isolution: "Get solution column W"; vtime Isolution:
    W := Transpose(Solution(Transpose(P), Transpose(UC)));

    // "SH:", SH; "W:", W;

    NC := HermiteForm(NC);
    if e eq 1 and Nrows(NC) gt 0 then
	M := NC[1, 1];
	if M ne 0 then
	    vprint Isolution: "Mod W by bottom entry", M;
	    for i := 1 to Nrows(W) do
		W[i, 1] := W[i, 1] mod M;
	    end for;
	end if;
    end if;

    InsertBlock(~X, SH, 1, 1);
    InsertBlock(~X, W, 1, s + 1);
    for i := s + 1 to k do
	X[i] := 0;
    end for;
    InsertBlock(~X, NC, s + 1, s + 1);

    vprint Isolution: "Final classical Hermite"; vtime Isolution:
    X := HermiteForm(X: Al := "Classical", InitialSort := false);

    return X;

end intrinsic;

intrinsic SingleSolutionTest(X::Mtrx, W::Mtrx) -> Mtrx
{true and solution V with V*X=W, or false if no solution}

    require BaseRing(X) cmpeq Z and BaseRing(W) cmpeq Z:
	"Matrices must be integral";

    n := Ncols(X);
    require Ncols(W) eq n: "RHS must have the same number of columns";

    B, U, N, Ypp, Y := LatticeBasisMatrix(X);
    vprint Isolution: "Get solution for B"; vtime Isolution:
    l, S := IsConsistent(B, W);
    if not l then
	return false, _;
    end if;

    m := Nrows(X);
    r := Nrows(S);
    s := Ncols(S);
// "S:", S; "B:", B; "U:", U; "Ypp:", Ypp; "s:", s; "Y:", Y;

    S *:= U;
    delete U;
// "New S:", S;
// "New S*Y:", S*Y;

    if Type(W) eq ModTupRngElt then
	V := RSpace(Z, m)!0;
	for j := 1 to Ncols(S) do
	    //V[Ypp[j]] := S[j];
	    V[j] := S[Ypp[j]];
	end for;
    else
	V := RMatrixSpace(Z, r, m)!0;
	for i := 1 to r do
	    for j := 1 to Ncols(S) do
		V[i, Ypp[j]] := S[i, j];
	    end for;
	end for;
    end if;

    return true, V;

end intrinsic;
