freeze;


CheckMSets:=function(G,m)
  L:=Subsets(Set(GSet(G)),m);
  L:=GSet(G,L);
  GG:=ActionImage(G,L);
  O:=Orbits(GG);
  return [#x:x in O], O;
end function;

intrinsic GaloisProof(K::FldNum, S::GaloisData: Level := 0) -> BoolElt
  {}
  return GaloisProof(DefiningPolynomial(K), S:Level := Level);
end intrinsic;

intrinsic GaloisProof(f::RngUPolElt, S::GaloisData:Level := 0) -> BoolElt
  {Verifies the descent steps of the Stauduhar method that were not proven during the computation.}

  if Level eq 0 then
    x := Position([y[2] : y in S`DescentChain], false);
    if x eq 0 then
      vprint GaloisGroup, 1: "All steps were proven";
      return true;
    end if;
    assert x ge 2;
    Level := x;
    G := S`DescentChain[x-1][1];
    H := S`DescentChain[#S`DescentChain][1];
  else
    if S`DescentChain[Level][2] then
      vprint GaloisGroup: "Step ", Level, " already checked";
      return true;
    end if;
    G := S`DescentChain[Level][1];
    x := Level-1;
    H := S`DescentChain[x][1];
  end if;

  n := Degree(f);
  m := 1;
  O := 1;
  repeat
    m +:= 1;
    mG, OG := CheckMSets(G, m);
    mH, O := CheckMSets(H, m);
  until mG ne mH;

  vprintf GaloisGroup, 1: "Groups can be distinguished on %o sets\n", m;
  vprintf GaloisGroup, 2: "for G we get: %o\n", mG;
  vprintf GaloisGroup, 2: "and for H   : %o\n", mH;

  use := false;
  for i in O do
    if i notin OG then
      if use cmpeq false or #i lt #use then
        use := i;
      end if;
    end if;
  end for;

  vprintf GaloisGroup, 3: "using orbit %o...\n", use;

  vprint GaloisGroup, 1: "Computing factor mod p...";
  vprint GaloisGroup, 2: "Roots...";
  vtime GaloisGroup, 2: r := [GaloisRoot(i, S: Prec := 1) : i in [1..n]];
  R := ResidueClassField(Universe(r));
  ChangeUniverse(~r, R);

  t := 0;
  vprint GaloisGroup, 1: "Computing MSetPol...", m;
  repeat
    vtime GaloisGroup, 1: F := MSetPolynomial(Evaluate(f, Polynomial([t, 1])), m);
    if IsSquarefree(Polynomial(R, F)) then
      break;
    end if;
    t +:= 1;
    if t gt 10 then
      error "Error: Too many transforms for MSetPol - maybe need different kind of them?";
    end if;
    vprint GaloisGroup, 2: "MSetPol not squarefree, trying substitution";
  until false;

  vprint GaloisGroup, 2: "Factors...";
  vtime GaloisGroup, 2: 
               lf := [ Polynomial([-&* [ r[i]-t : i in u], 1]) : u in use];
  vprint GaloisGroup, 2: "Product...";
  vtime GaloisGroup, 2: q1 := &* lf;

  vprint GaloisGroup, 1: "Computing cofactor...";
  vtime GaloisGroup, 2: f, q2 := IsDivisibleBy(Polynomial(R, F), q1);
  assert f;

  vprint GaloisGroup, 1: "Computing bounds...";
  B := 2^#O*Sqrt(Degree(F)+1)*Maximum([Abs(x) : x in Eltseq(F)]);
  B := Ceiling(B);

  prec := Ilog(S`Prime, 2*B);
  vprint GaloisGroup, 1: "using precision bound ", prec, "for lifting";
  R := BaseRing(S`Ring);
  R := ChangePrecision(R, prec);
  FL := Polynomial(R, F);
  fl, qq1 := IsCoercible(Parent(FL), q1);
  if not fl then
    vprint GaloisGroup, 1: "Factorisation is not over BaseRing";
    return false;
  end if;
  fl, qq2 := IsCoercible(Parent(FL), q2);
  if not fl then
    vprint GaloisGroup, 1: "Factorisation is not over BaseRing";
    return false;
  end if;
  L := [ qq1, qq2];

  vprint GaloisGroup, 1: "Lifting...";
  vtime GaloisGroup, 1: L := HenselLift(FL, L);
  q1 := L[1];

  vprint GaloisGroup, 2: "Checking coefficients...";
  q1 := Eltseq(q1);
  Q1 := [];
  for i in q1 do
    fl, p := IsInt(S`Ring!BaseRing(S`Ring)!i, B, S);
    if not fl then
      vprint GaloisGroup, 1: "Coefficient too large...";
      return false;
    end if;
    Append(~Q1, p);
  end for;

  vprint GaloisGroup, 1: "Final division...";
  Q1 := Polynomial(Q1);
  vtime GaloisGroup, 1: fl := IsDivisibleBy(F, Q1);
 
  S`DescentChain[Level][2] := true;

  return fl;
end intrinsic;    

    
